#include <Coin_C_defines.h>
#include <CoinPackedMatrix.hpp>
#include <CoinMessageHandler.hpp>
#include <ClpSimplex.hpp>
#include <ClpInterior.hpp>
#include <ClpCholeskyBase.hpp>

extern "C"
{
#include <stack-c.h>
}
#undef CreateVar
#define CreateVar(n,ct,mx,nx,lx) if(! C2F(createvar)((c_local=n,&c_local),ct,mx,nx,lx, 1L)) \
    { return 0;  }

#define dstk(x) (((double *) C2F(stack).Stk) + x-1 )

#define DEBUG 1

/*
  Disclaimer : Almost no error checks (relies on checks in clp.m). Horrible C-code (I'm a MATLAB coder....)
  Author     : Johan Löfberg, ETH Zürich, loefberg@control.ee.ethz.ch
*/

class DerivedHandler : public CoinMessageHandler 
{
public:
  virtual int print() ;
};

int DerivedHandler::print()
{
  sciprint("%s\n",messageBuffer());
  return 0;
}

int find_label(char **LabelList, int nblabels, char * LabelToFind)
{
  int Pos = -1, i;

  for(i=0; i<nblabels; i++)
    {
      if (strcmp(LabelList[i],LabelToFind)==0)
	{
	  Pos = i;
	  return Pos;
	} 
    } 
  return Pos;
} 



/*
  In Clp_C_Interface.h

  COINLIBAPI int COINLINKAGE Clp_getIterationCount(Clp_Simplex * model); // Are there a numerical difficulties? 
  COINLIBAPI int COINLINKAGE Clp_isAbandoned(Clp_Simplex * model);       // Is optimality proven?
  COINLIBAPI int COINLINKAGE Clp_isProvenOptimal(Clp_Simplex * model);   // Is primal infeasiblity proven? 
  COINLIBAPI int COINLINKAGE Clp_isProvenPrimalInfeasible(Clp_Simplex * model); // Is dual infeasiblity proven?
  COINLIBAPI int COINLINKAGE Clp_isProvenDualInfeasible(Clp_Simplex * model);   // Is the given primal objective limit reached? 
  COINLIBAPI int COINLINKAGE Clp_isPrimalObjectiveLimitReached(Clp_Simplex * model); // Is the given dual objective limit reached?
  COINLIBAPI int COINLINKAGE Clp_isDualObjectiveLimitReached(Clp_Simplex * model) ;  // Iteration limit reached?
  COINLIBAPI int COINLINKAGE Clp_isIterationLimitReached(Clp_Simplex * model);       // Direction of optimization (1 - minimize, -1 - maximize, 0 - ignore 
  COINLIBAPI double COINLINKAGE Clp_getObjSense(Clp_Simplex * model);                // Primal row solution 
  COINLIBAPI const double * COINLINKAGE Clp_getRowActivity(Clp_Simplex * model);     // Primal column solution 
  COINLIBAPI const double * COINLINKAGE Clp_getRowPrice(Clp_Simplex * model);        
  COINLIBAPI const double * COINLINKAGE Clp_getReducedCost(Clp_Simplex * model);     // Reduced costs 
  COINLIBAPI const double* COINLINKAGE Clp_getRowLower(Clp_Simplex * model);         // Row lower 
  COINLIBAPI const double* COINLINKAGE Clp_getRowUpper(Clp_Simplex * model);         // Row upper
  COINLIBAPI const double * COINLINKAGE Clp_getColLower(Clp_Simplex * model);        // Column Lower
  COINLIBAPI const double * COINLINKAGE Clp_getColUpper(Clp_Simplex * model);        // Column Upper
  COINLIBAPI double COINLINKAGE Clp_getObjValue(Clp_Simplex * model);                // Objective value
  COINLIBAPI double COINLINKAGE Clp_getSmallElementValue(Clp_Simplex * model);       // Small element value - elements less than this set to zero, default is 1.0e-20
  COINLIBAPI void COINLINKAGE Clp_setSmallElementValue(Clp_Simplex * model,double value);

*/

#define A_IN        1
#define C_IN        2
#define B_IN        3
#define NEQ         4
#define LOWER       5
#define UPPER       6
#define TYPE        7
#define Q_IN        8
#define PARAM       9
#define LASTPARAM   9

extern "C" int C2F(sciclp)(char * fname);

int C2F(sciclp)(char * fname)
{
  ClpSimplex     * modelByColumn   = NULL;
  ClpInterior    * modelPrimalDual = NULL;
  DerivedHandler * printer         = NULL;
  CoinPackedMatrix A_matrix, Q_matrix;

	    
  double *primal, *dual;
    
  int ncols, nrows, i, j, nz;
  int count = 0;
  
  if (Rhs<LASTPARAM) Scierror(999,"12 inputs required in call to sciclp. Bug in clp.sci ?...\n");
		
  /* Get pointers to input */

  int n_a,     m_a,     l_a;
  int n_c,     m_c,     l_c;
  int n_b,     m_b,     l_b;
  int n_neq,   m_neq,   l_neq;
  int n_upper, m_upper, l_upper;
  int n_lower, m_lower, l_lower;
  int n_type,  m_type,  l_type;
  int n_q,     m_q,     l_q;

  GetRhsVar(C_IN,  "d", &n_c,     &m_c,     &l_c);   // double
  GetRhsVar(B_IN,  "d", &n_b,     &m_b,     &l_b);   // double
  GetRhsVar(NEQ,   "i", &n_neq,   &m_neq,   &l_neq);   // int
  GetRhsVar(LOWER, "d", &n_lower, &m_lower, &l_lower); // double
  GetRhsVar(UPPER, "d", &n_upper, &m_upper, &l_upper); // double
  GetRhsVar(TYPE,  "c", &n_type,  &m_type,  &l_type);  // char *

  ncols = n_c * m_c; /* Length of c == number of columns */
  nrows = n_b * m_b; /* length of b == number of rows    */

#ifdef DEBUG
  printf("DEBUG: A           size: n = %d m= %d\n", n_a,  m_a);
  printf("DEBUG: c           size: n = %d m= %d\n", n_c,  m_c);
  printf("DEBUG: b           size: n = %d m= %d\n", n_b,  m_b);
  printf("DEBUG: neq         size: n = %d m= %d neq value = %d\n", n_neq,m_neq, *istk(l_neq));
  printf("DEBUG: lower       size: n = %d m= %d\n", n_lower, m_lower);
  printf("DEBUG: upper       size: n = %d m= %d\n", n_upper, m_upper);
  printf("DEBUG: type        size: n = %d m= %d type = %s\n", n_type, m_type, cstk(l_type));
  printf("DEBUG: Q           size: n = %d m= %d\n", n_q, m_q);
  printf("DEBUG: nrows = %d\n", nrows);
  printf("DEBUG: ncols = %d\n", ncols);

  for(i=0;i<ncols;i++)
    {
      printf("DEBUG: Var[%d]: type = %s, lower = %f, upper = %f\n", i, cstk(l_type+i), *dstk(l_lower+i), *dstk(l_upper+i));
    }
  for(i=0;i<ncols;i++)
    {
      printf("DEBUG: obj[%d] = %f\n",i,*dstk(l_c+i));
    }
  for(i=0;i<nrows;i++)
    {
      printf("DEBUG: b[%d] = %f\n", i, *dstk(l_b+i));
    }
#endif

  A_matrix.setDimensions(nrows,ncols);
  
  if(GetType(A_IN)!=5) // type(A)==5 -> A is a sparse matrix
    {
#ifdef DEBUG
      printf("DEBUG: A_IN is not sparse\n");
#endif
      
      GetRhsVar(A_IN, "d", &m_a, &n_a, &l_a); // get the matrix
      
      if (dstk(l_a)==NULL) Scierror(999,"sciclp: invalid value of matrix a");
      
      for(i=0; i<m_a; i++)
	{
	  for(j=0; j<n_a; j++)
	    {
	      if (*dstk(l_a+i*n_a+j) != 0) A_matrix.modifyCoefficient(i,j,*dstk(l_a+i*n_a+j));
	    }
	}
     }
   else
     {
#ifdef DEBUG
       printf("DEBUG: A_IN is sparse\n");
#endif

       SciSparse S_A;
       
       GetRhsVar(A_IN, "s", &m_a, &n_a, &S_A);

       nz = S_A.nel;
       
       count = 0;
       for(i=0;i<S_A.m;i++)
	 {
	   if (S_A.mnel[i]!=0) 
	     {
	       for(j=0;j<S_A.mnel[i];j++)
		 {
		   count++;
		   A_matrix.modifyCoefficient(i,S_A.icol[count-1],S_A.R[count-1]);
		 }
	     }
	 }
     }

  // We must sets the upper and lower bounds using clp.sci function !

  // the first neq constraints are equality constraints
  double * lhs = NULL;

  lhs = (double *)malloc(nrows*sizeof(double));
  for (i=0; i<nrows; i++) lhs[i] = -COIN_DBL_MAX;
  
  for (i=0; i<*istk(l_neq); i++) lhs[i] = *dstk(l_b+i);

  /////////////////
  // Get options //
  /////////////////

#ifdef DEBUG
  printf("DEBUG: get options\n");
#endif

  // Default settings
  // optimisation direction: 1 - minimize, -1 - maximize, 0 - ignore
  int     solverchoice  = 1, maxnumiterations = 99999999, loglevel = 0, optim_dir = 1;
  double  maxnumseconds = 3600.0, primaltolerance = 1e-7, dualtolerance = 1e-7;
  int     m_param, n_param, l_param;
  int     m_tmp,   n_tmp,   l_tmp;
  int     m_label, n_label, pos_label;
  char ** LabelList = NULL;

  GetRhsVar(PARAM, "m", &m_param, &n_param, &l_param);
  GetListRhsVar(PARAM, 1, "S", &m_label, &n_label, &LabelList);

  if (strcmp(LabelList[0],"plist") != 0) 
    {
      Scierror(999,"Argument %d is not a plist\r\n", 12);
      return 0;
    }
  else
    {
      // maxnumiterations option
      pos_label = find_label(LabelList, m_param, "maxnumiterations");
      if (pos_label!=-1) 
	{
	  GetListRhsVar(PARAM, pos_label+1, "i", &m_tmp, &n_tmp, &l_tmp);
	  maxnumiterations = *istk(l_tmp);
#ifdef DEBUG
	  printf("DEBUG: maxnumiterations = %d\n", maxnumiterations);
#endif
	}
      
      // maxnumseconds option
      pos_label = find_label(LabelList, m_param, "maxnumseconds");
      if (pos_label!=-1) 
	{
	  GetListRhsVar(PARAM, pos_label+1, "d", &m_tmp, &n_tmp, &l_tmp);
	  maxnumseconds = *dstk(l_tmp);
#ifdef DEBUG
	  printf("DEBUG: maxnumseconds = %f\n", maxnumseconds);
#endif
	}
      
      // primaltolerance option
      pos_label = find_label(LabelList, m_param, "primaltolerance");
      if (pos_label!=-1) 
	{
	  GetListRhsVar(PARAM, pos_label+1, "d", &m_tmp, &n_tmp, &l_tmp);
	  primaltolerance = *dstk(l_tmp);
#ifdef DEBUG
	  printf("DEBUG: primaltolerance = %f\n", primaltolerance);
#endif
	}

      // dualtolerance option
      pos_label = find_label(LabelList, m_param, "dualtolerance");
      if (pos_label!=-1) 
	{
	  GetListRhsVar(PARAM, pos_label+1, "d", &m_tmp, &n_tmp, &l_tmp);
	  dualtolerance = *dstk(l_tmp);
#ifdef DEBUG
	  printf("DEBUG: dualtolerance = %f\n", dualtolerance);
#endif
	}

      // verbose option
      pos_label = find_label(LabelList, m_param, "verbose");
      if (pos_label!=-1) 
	{
	  GetListRhsVar(PARAM, pos_label+1, "i", &m_tmp, &n_tmp, &l_tmp);
	  loglevel = *istk(l_tmp);
#ifdef DEBUG
	  printf("DEBUG: loglevel = %d\n", loglevel);
#endif
	}

      // solver option
      pos_label = find_label(LabelList, m_param, "solver");
      if (pos_label!=-1) 
	{
	  GetListRhsVar(PARAM, pos_label+1, "i", &m_tmp, &n_tmp, &l_tmp);
	  solverchoice = *istk(l_tmp);
#ifdef DEBUG
	  printf("DEBUG: solverchoice = %d\n", solverchoice);
#endif
	}

      // optim_dir option
      pos_label = find_label(LabelList, m_param, "optim_dir");
      if (pos_label!=-1) 
	{
	  GetListRhsVar(PARAM, pos_label+1, "i", &m_tmp, &n_tmp, &l_tmp);
	  optim_dir = *istk(l_tmp);
#ifdef DEBUG
	  printf("DEBUG: optim_dir = %d\n", optim_dir);
#endif
	}
    }
	
  switch (solverchoice)
    {
    default:
      modelByColumn = new ClpSimplex();
      modelByColumn->loadProblem(A_matrix,dstk(l_lower),dstk(l_upper),dstk(l_c),lhs,dstk(l_b));
      if (n_type*m_type!=0)
	{
	  for(i=0;i<(int)strlen(cstk(l_type)); i++)
	    {
	      if (*cstk(l_type+i)=='I') modelByColumn->setInteger(i);
	      else                      modelByColumn->setContinuous(i);
	    }
	}
      break;
    case 3:
      modelPrimalDual = new ClpInterior();
      modelPrimalDual->loadProblem(A_matrix,dstk(l_lower),dstk(l_upper),dstk(l_c),lhs,dstk(l_b));			
      if (n_type*m_type!=0)
	{
	  for(i=0;i<(int)strlen(cstk(l_type)); i++)
	    {
	      if (*cstk(l_type+i)=='I') modelPrimalDual->setInteger(i);
	      else                      modelPrimalDual->setContinuous(i);
	    }
	}
      break;
    }
		
  ////////////////////////
  // Any quadratic part //
  ////////////////////////

  Q_matrix.setDimensions(nrows,ncols);
  
  if(GetType(Q_IN)!=5) // type(Q)==5 -> A is a sparse matrix
    {
#ifdef DEBUG
      printf("DEBUG: Q_IN is not sparse\n");
#endif
      
      GetRhsVar(Q_IN, "d", &m_q, &n_q, &l_q); // get the matrix
      
      if (dstk(l_q)==NULL) Scierror(999,"sciclp: invalid value of matrix q");
      
      for(i=0; i<m_q; i++)
	{
	  for(j=0; j<n_q; j++)
	    {
	      if (*dstk(l_q+i*n_q+j) != 0) Q_matrix.modifyCoefficient(i,j,*dstk(l_q+i*n_q+j));
	    }
	}
     }
   else
     {
#ifdef DEBUG
       printf("DEBUG: Q_IN is sparse\n");
#endif

       SciSparse S_Q;
       
       GetRhsVar(Q_IN, "s", &m_q, &n_q, &S_Q);

       nz = S_Q.nel;
       
       count = 0;
       for(i=0;i<S_Q.m;i++)
	 {
	   if (S_Q.mnel[i]!=0) 
	     {
	       for(j=0;j<S_Q.mnel[i];j++)
		 {
		   count++;
		   Q_matrix.modifyCoefficient(i,S_Q.icol[count-1],S_Q.R[count-1]);
		 }
	     }
	 }
     }

  if (nz>0)
    {	
      if (solverchoice==2) solverchoice = 1;
      
      switch (solverchoice)
	{
	default:
	  modelByColumn->loadQuadraticObjective(Q_matrix);
	  break;
	case 3:
	  modelPrimalDual->loadQuadraticObjective(Q_matrix);
	}					
    }	
  
  ///////////////////////////////
  // Enable printing in Scilab //
  ///////////////////////////////

  printer = new DerivedHandler(); // assumed open	
  printer->setLogLevel(loglevel);		 
		
  if (solverchoice == 3)
    {		
      modelPrimalDual->setMaximumIterations(maxnumiterations);
      modelPrimalDual->setMaximumSeconds(maxnumseconds);
      modelPrimalDual->setPrimalTolerance(primaltolerance);
      modelPrimalDual->setDualTolerance(dualtolerance);	
      modelPrimalDual->passInMessageHandler(printer);
      modelPrimalDual->setOptimizationDirection(optim_dir);
    }
  else
    {
      modelByColumn->setMaximumIterations(maxnumiterations);
      modelByColumn->setMaximumSeconds(maxnumseconds);
      modelByColumn->setPrimalTolerance(primaltolerance);
      modelByColumn->setDualTolerance(dualtolerance);				
      modelByColumn->passInMessageHandler(printer);	
      modelByColumn->setOptimizationDirection(optim_dir);
    }
  
  switch (solverchoice)
    {
    default:
    case 1:			
      modelByColumn->primal();		
      break;
    case 2:			
      modelByColumn->dual();
      break;
    case 3:
      ClpCholeskyBase * cholesky = new ClpCholeskyBase();			
      cholesky->setKKT(true);		
      modelPrimalDual->setCholesky(cholesky);	
					
      if (modelPrimalDual->primalDual()) sciprint("Failed\n");
      break;
    }

  //////////////////////////////
  // Allocate for return data //
  //////////////////////////////

#ifdef DEBUG
  printf("DEBUG: allocating data\n");
#endif

  int n_x,          m_x,          l_x;          // double
  int n_lambda,     m_lambda,     l_lambda;     // double
  int n_returncode, m_returncode, l_returncode; // int

  n_x = 1;          m_x = ncols;  // a row vector
  n_lambda = nrows; m_lambda = 1; // a column vector
  n_returncode = 1; m_returncode = 1;

  CreateVar(1, "d", &m_x,          &n_x,          &l_x);
  CreateVar(2, "d", &m_lambda,     &n_lambda,     &l_lambda);
  CreateVar(3, "i", &m_returncode, &n_returncode, &l_returncode);

#ifdef DEBUG
  printf("DEBUG: solving\n");
#endif

  if (solverchoice==3)
    {	
      primal              = modelPrimalDual->primalColumnSolution();
      dual                = modelPrimalDual->dualRowSolution();
      *istk(l_returncode) = modelPrimalDual->status();			
    }
  else
    {	
      primal              = modelByColumn->primalColumnSolution();
      dual                = modelByColumn->dualRowSolution();
      *istk(l_returncode) = modelByColumn->status();		
    }
  
  //
  // Status of problem:
  //  -1 - unknown e.g. before solve or if postSolve says not optimal
  //   0 - optimal
  //   1 - primal infeasible
  //   2 - dual infeasible
  //   3 - stopped on iterations or time
  //   4 - stopped due to errors
  //   5 - stopped by event handler (virtual int ClpEventHandler::event())
  //

  /////////////////////////////////
  // Copy solutions if available //
  /////////////////////////////////

#ifdef DEBUG
  printf("DEBUG: returning data\n");
#endif

  if (primal!=NULL) memcpy(dstk(l_x),      primal, ncols*sizeof(double));
  if (dual!=NULL)   memcpy(dstk(l_lambda), dual,   nrows*sizeof(double));

  LhsVar(1) = 1;
  LhsVar(2) = 2;
  LhsVar(3) = 3;

  PutLhsVar();

  //////////////////////////////
  // Delete allocated objects //
  //////////////////////////////

  if (modelByColumn!= NULL)   delete(modelByColumn);
  if (modelPrimalDual!= NULL) delete(modelPrimalDual);
  if (printer!= NULL)         delete(printer);	
  if (lhs!=NULL)              free(lhs);

  return 0;
}




